c*************************************************************************
c                                                                        *
c   PBFIT - Mike Towler (March 1997)                                     *
c                                                                        *
c   Do a non-linear fit of a pseudopotential given on a radial grid      *
c   using a modified levenberg-marquardt algorithm and a basis of        *
c   gaussian functions * 1/r^n.                                          *
c                                                                        *
c   Version 1: March 1997                                                *
c   Version 2: October 1999 - introduced a constraint to avoid numerical *
c              instabilities which led to fluctuations in the fitted     *
c              pseudo near the nucleus under certain circumstances.      *
c   Version 3: November 1999 - allowed more control of fit from control  *
c              file plus a stack of f90 modifications.                   *
c                                                                        *
c   Notes:                                                               *
c   (1) accuracy of fit very sensitive to starting guess  - rerunning    *
c       with different initial coefficients and exponents is usually     *
c       a good idea - see shell script                                   *
c                                                                        *
c   INPUT:                                                               *
c                                                                        *
c   standard input: single pseudopotential component on a radial grid    *
c                   (in cambridge qmc x_pp.data format)                  *
c   'control'     : gives various parameters required for the run        *
c                   (see comments in example file for details)           *
c                                                                        *
c   OUTPUT:                                                              *
c                                                                        *
c   standard output : log file containing parameters and fitting data at *
c                     each iteration.                                    *
c                                                                        *
c   'PLOT'         : xmgr plot of fit to r^n*V (au)                      *
c   'PLOTV'        : xmgr plot of fit to V derived from PLOT (au)        *
c                                                                        *
c*************************************************************************

      MODULE pfit_data
       INTEGER na,ma,ndata,numpower,numbf,offset,rpower
       DOUBLE PRECISION,DIMENSION(:),ALLOCATABLE :: rn,nk
       DOUBLE PRECISION fixed_coeff,znuc
       LOGICAL constraint,local
      END MODULE pfit_data

      PROGRAM pbfit
      USE pfit_data
      IMPLICIT NONE
      INTEGER i,j,k,n,lwa,info,inonlocal,ialloc,irydberg,it,maxit,
     & iconstraint
      INTEGER,DIMENSION(:),ALLOCATABLE :: ipvt
      DOUBLE PRECISION pcoeff,amin,amax,tol,mdiff,chisq,rsq,alpha,b,c
      DOUBLE PRECISION,DIMENSION(:),ALLOCATABLE :: as,asp
      DOUBLE PRECISION,DIMENSION(:),ALLOCATABLE :: fvec,r,wa,ys,yfits
      DOUBLE PRECISION,DIMENSION(:,:),ALLOCATABLE :: fjac
      CHARACTER*72 dum
      LOGICAL rydberg,istop
      EXTERNAL fgauss
c
c*************************************************************************
c
      write(6,*)'######  ######  #######   ###   #######'
      write(6,*)'#     # #     # #          #       #'
      write(6,*)'#     # #     # #          #       #'
      write(6,*)'######  ######  #####      #       #'
      write(6,*)'#       #     # #          #       #'
      write(6,*)'#       #     # #          #       #'
      write(6,*)'#       ######  #         ###      #'
      write(6,*)' '
      write(6,*)'Gaussian pseudopotential fitter log file'
      write(6,*)'========================================'
      write(6,*)
c.....read control file
      open(unit=10,file='control',status='old')
      read(10,'(a)')dum
      read(10,'(a)')dum
      read(10,'(a)')dum
      read(10,'(a)')dum
      read(10,*,err=1)znuc
      if(znuc.lt.1.d0.and.znuc.gt.98.d0)then
       write(6,*)'I think znuc is probably out of range.'
       istop=.true.
      endif
      read(10,'(a)')dum
      read(10,*,err=1)inonlocal
      if(inonlocal.ne.0.and.inonlocal.ne.1)then
       write(6,*)'NONLOCAL out of range.'
       istop=.true.
      endif
      local=inonlocal.eq.0
      read(10,'(a)')dum
      read(10,*,err=1)rpower
      if(rpower.lt.0.or.rpower.gt.2)then
       write(6,*)'RPOWER out of range.'
       istop=.true.
      endif
      read(10,'(a)')dum
      read(10,*,err=1)iconstraint
      if(iconstraint.ne.0.and.iconstraint.ne.1)then
       write(6,*)'CONSTRAINT out of range.'
       istop=.true.
      endif
      constraint=iconstraint.eq.1
      if(constraint.and.inonlocal.eq.1)then
       write(6,*)'Don''t use constraint when fitting nonlocal V1-V2'
       istop=.true.
      endif
      read(10,'(a)')dum
      read(10,*,err=1)irydberg
      rydberg=irydberg.eq.1
      if(irydberg.ne.1.and.irydberg.ne.2)then
       write(6,*)'RYDBERG out of range.'
       istop=.true.
      endif
      read(10,'(a)')dum
      read(10,*,err=1)amin
      read(10,*,err=1)amax
      if(amin.lt.0.001)then
       write(6,*)'AMIN is too small'
       istop=.true.
      endif
      if(amin.ge.amax)then
       write(6,*)'AMIN is bigger than AMAX'
       istop=.true.
      endif
      read(10,'(a)')dum
      read(10,'(a)')dum
      read(10,*,err=1)pcoeff
      read(10,'(a)')dum
      read(10,*,err=1)numpower
      if(numpower.gt.9)then
       write(6,*)'NUMPOWER out of range in control file (max: 10)'
       istop=.true.
      endif
      allocate(nk(numpower),stat=ialloc)
      read(10,'(a)')dum
      read(10,'(a)')dum
      read(10,*,err=1)(nk(i),i=1,numpower)
      if (any(nk(:).lt.-1).or.any(nk(:).gt.7))then
       write(6,*)'Power of r out of range in control file: use -1 to 7'
       istop=.true.
      endif
      do i=1,numpower-1
       if(nk(i).lt.nk(i+1))then
        write(6,*)
     &  'Write powers of r in decreasing order in control file'
        istop=.true.
        goto 10
       endif
      enddo
10    read(10,'(a)')dum
      read(10,*,err=1)na
      ma=2*numpower*na ! number of variable parameters
      if(constraint)ma=ma-1
      if(istop)then
       write(6,*) "Quitting.."
       stop
      endif
      goto 2
1     write(6,*)'some problem occurred reading the control file.' 
      stop
c.....read data file
2     read(5,'(a)')dum 
      read(5,'(a)')dum 
      read(5,*)ndata
c.....ignore r=0 to avoid numerical problems
      ndata=ndata-1
      lwa=5*ma+ndata
      ialloc=0
      allocate(ipvt(ma),as(ma),asp(ma),fvec(ndata),r(ndata),
     & ys(ndata),yfits(ndata),wa(lwa),fjac(ndata,ma),
     & rn(na*numpower),stat=ialloc)
      if(ialloc.ne.0)then
       write(6,*)'allocation problem'
       stop
      endif
      read(5,'(a)')dum 
      read(5,'(a)')dum 
      do i=1,ndata
       read(5,*)r(i)
      enddo
      read(5,'(a)')dum 
      read(5,'(a)')dum 
      do i=1,ndata
       read(5,*)ys(i)
      enddo
c.....divide data by two if in rydbergs
      
      if(rydberg)ys=ys*0.5d0

c.....initialize various things
      
      n=0
      do i=1,numpower
       do j=1,na
        n=n+1
        rn(n)=nk(i)+rpower
       enddo
      enddo
      offset=numpower*na ! last exponent position in parameter vector
      numbf=offset ! number of r^n*exp basis functions
      if(constraint)numbf=numbf-1

c.....Initialize parameter vector
c.....Initial exponents even-tempered from amin to amax

      call guess(pcoeff,as,amax,amin)
      
c.....Generate initial best fit coefficients without varying
c.....the exponents (helps with numerical stability)
 
      call coeffs(r,ys,as)
 
c.....Initial convergence criterion
c.....Termination of an iteration occurs when relative error in 
c.....sum of squared deviations is at most tol, or the relative error 
c.....between parameter vector and the solution vector is at most tol.
c.....In this version, this initial tolerance will be gradually 
c.....lowered until the fit can't be improved any more

      tol=1.d-1
      maxit=15
      it=0
2000  it=it+1
      if(it.gt.maxit)goto 3000
      write(6,*)'Begin iteration with TOL ', tol
      write(6,*)
      call lmder1(r,ys,fgauss,ndata,ma,as,fvec,fjac,ndata,tol,info,
     &            ipvt,wa,lwa)

      if(info.eq.1.or.info.eq.2.or.info.eq.3)then
       if(constraint)then
        k=offset+na*(numpower-1)
        fixed_coeff=znuc
        do i=1,na-1
         k=k+1
         fixed_coeff=fixed_coeff-as(k)
        end do
       endif

       write(6,*)'SUCCESSFUL FIT.' 
       call reason(info,tol)
       mdiff=0.d0 ; chisq=0.d0
       do n=1,ndata
        yfits(n)=0.d0
        rsq=r(n)**2.d0
        do j=1,numbf
         i=offset+j
         alpha=-(as(j)**2.d0)
         yfits(n)=yfits(n)+as(i)*r(n)**rn(j)*dexp(alpha*rsq)
        enddo
        if(constraint)then
         yfits(n)=yfits(n)+fixed_coeff*r(n)**rn(offset)*
     &   dexp(-(as(offset)**2.d0)*rsq)
        endif
        if(local)yfits(n)=yfits(n)-znuc*r(n)**rn(offset)
        chisq=chisq+(ys(n)-yfits(n))**2.d0
        mdiff=max(abs(ys(n)-yfits(n)),mdiff)
       enddo ! grid points

       write(6,*)'SUM OF SQUARED DEVIATIONS (Y-YFIT)^2: ',chisq
       write(6,*)'MAXIMUM DIFFERENCE Y-YFIT: ',mdiff
       write(6,*)'GAUSSIAN PARAMETERS:'
       write(6,999)
999    format(5x,'coefficient',15x,'exponent',15x,'power of r')

       do j=1,numbf
        i=offset+j
        write(6,*)as(i),as(j)**2.d0,int(rn(j)-rpower)
       enddo
       if(constraint)write(6,*)fixed_coeff,as(offset)**2.d0,
     & int(rn(offset)-rpower)
       asp(1:ma)=as(1:ma)
       write(6,*)
       tol=tol/10.d0
       goto 2000
      else
       write(6,*)'Unable to find satisfactory fit.' 
       write(6,*)'Stop condition:' 
       call errmdt(info,tol)
       write(6,*)
       if(constraint)then
        k=offset+na*(numpower-1)
        write(6,*)'fixed_coeff1',fixed_coeff
        fixed_coeff=znuc
        do i=1,na-1
         k=k+1
         fixed_coeff=fixed_coeff-asp(k)
        end do
       endif
      endif
c     iteration
      write(6,*)

c.....plot the fit to data on grid (i.e. r*V or whatever) 
3000  open(unit=22,file='PLOT',status='unknown')
! X,Y1,Y2,Y3.. format in xmgr doesn't support comments (!)
c      write(22,*)'r y yfit diff' 
      do i=1,ndata
       if(dabs(ys(i)).lt.(1.d-99))ys(i)=0.d0
       if(dabs(yfits(i)).lt.(1.d-99))yfits(i)=0.d0
      enddo
      do i=1,ndata
       write(22,100)r(i),ys(i),yfits(i),ys(i)-yfits(i)
      enddo
100   format(1X,4(1PE18.10))
      close(22)
c.....If we haven't been fitting to r^0*V, then
c.....plot the fit to actual pseudo (i.e. V) derived from r^n*V fit
c.....(which is what GAUSSIAN/CRYSTAL actually use)
      if(rpower.ne.0)then
      open(unit=22,file='PLOTV',status='unknown')
! X,Y1,Y2,Y3.. format in xmgr doesn't support comments (!)
c      write(22,*)'r y/r^n yfit/r^n diff' 
       do i=1,ndata
        b=ys(i)*(1.d0/r(i)**rpower)
        c=yfits(i)*(1.d0/r(i)**rpower)
        if(dabs(b).lt.(1.d-99))b=0.d0
        if(dabs(c).lt.(1.d-99))c=0.d0
        write(22,100)r(i),b,c,b-c
       enddo
       close(22)
      endif

c.....write out final parameters
      write(6,*)
      write(6,*)'No further improvement possible'
      write(6,*)
      write(6,*)'Final results (gaussian98 format):'
      write(6,*)
c.....for Gaussian we need to include the Jacobian of r^2 in the
c.....powers of r
      write(6,1000)
1000  format(5x,'power of r + 2',4x,'exponent',13x,'coefficient')
      do j=1,numbf
       i=offset+j
       write(6,*)(int(rn(j)-rpower)+2),asp(j)**2.d0,asp(i)
      enddo
      if(constraint)write(6,*)(int(rn(offset)-rpower)+2),
     & asp(offset)**2.d0,fixed_coeff
      write(6,*)
      write(6,*)'Final results (crystal format):'
      write(6,*)
      write(6,'(t1,''INPUT'')')
      write(6,*)znuc,'a b c d e'
      do j=1,numbf
       i=offset+j
       write(6,*)asp(j)**2.d0,asp(i),int(rn(j)-rpower)
      enddo
      if(constraint)write(6,*)asp(offset)**2.d0,fixed_coeff,
     & int(rn(offset)-rpower)
      stop
      END PROGRAM pbfit
   
      SUBROUTINE guess(pcoeff,as,amax,amin)
      USE pfit_data
      IMPLICIT NONE
      INTEGER i,j,k,n
      DOUBLE PRECISION as(ma),amax,amin,a1,ratio,expon(na),pcoeff
c.....initialize coefficients
      as(offset+1:ma)=pcoeff
c.....define set of even-tempered exponents
      a1=(dlog10(amax/amin))/(dfloat(na-1))
      ratio=10.d0**a1
      do i=1,na
       expon(i)=amin*ratio**dfloat(i-1)
      enddo
c.....and shove them into the vector of variable parameters
      n=0
      do i=1,numpower
       do j=1,na
        n=n+1
        as(n)=dsqrt(expon(j))
       enddo
      enddo
      return
      END SUBROUTINE guess

      SUBROUTINE fgauss(r,y,m,n,x,fvec,fjac,ldfjac,iflag)
c m=ndata, n=ma
      USE pfit_data
      IMPLICIT NONE
      INTEGER m,n,ldfjac,iflag
      DOUBLE PRECISION y(m),r(m),fvec(m),fjac(ldfjac,n),x(n)
      INTEGER i,j,k,l
      DOUBLE PRECISION yfit,arg,ex,dist,fac
      if(constraint)then
       k=offset+na*(numpower-1)
       fixed_coeff=znuc
       do i=1,na-1
        k=k+1
        fixed_coeff=fixed_coeff-x(k)
       end do
      endif
      if(iflag.eq.1)then
       do l=1,m ! loop over grid points
        dist=r(l)
        yfit=0.d0	
        do j=1,numbf
	 i=offset+j
         arg=x(j)**2.d0*dist**2.d0
         ex=dexp(-arg)*dist**rn(j)
         fac=x(i)*ex
         yfit=yfit+fac 
        end do     
	if(constraint)then 
         arg=x(offset)**2.d0*dist**2.d0
         ex=dexp(-arg)*dist**rn(offset)
         fac=fixed_coeff*ex
         yfit=yfit+fac
	endif
	if(local)yfit=yfit-(znuc*dist**rn(offset))
	fvec(l)=y(l)-yfit
       end do ! grid points
      elseif(iflag.eq.2)then ! evaluate derivatives
       do l=1,m ! grid points
        dist=r(l)
        do j=1,numbf
	 i=offset+j
         arg=x(j)**2.d0*dist**2.d0
         ex=dexp(-arg)*dist**rn(j)
         fac=x(i)*ex
         fjac(l,i)=-ex 	
	 fjac(l,j)=2.d0*x(j)*dist**2.d0*fac
        end do     
	if(constraint)then
         arg=x(offset)**2.d0*dist**2.d0
         ex=dexp(-arg)*dist**rn(offset)
	 fac=fixed_coeff*ex
         fjac(l,offset)=2.d0*x(offset)*dist**2.d0*fac
         k=k+2
	endif
       enddo
      elseif(iflag.eq.0)then
       write(6,*)'Go away.'
       write(6,*)'Mike has not told me how to print yet.'
      else
       write(6,*)'Invalid value of iflag in fgauss.'     
       stop
      endif
      return
      END SUBROUTINE fgauss

      SUBROUTINE fgauss2(x,a,y,dyda)
      USE pfit_data
      IMPLICIT NONE
      INTEGER i,j,k
      DOUBLE PRECISION x,y,a(ma),dyda(ma)
      DOUBLE PRECISION arg,ex,fac
      y=0.d0
      if(constraint)then
       k=offset+na*(numpower-1)
       fixed_coeff=znuc
       do i=1,na-1
        k=k+1
        fixed_coeff=fixed_coeff-a(k)
       end do
      endif
      do j=1,numbf
       i=offset+j
       arg=a(j)**2.d0*x**2.d0
       ex=dexp(-arg)*x**rn(j)
       fac=a(i)*ex
       y=y+fac
       dyda(i)=ex
       dyda(j)=(-2.d0)*a(j)*x**2.d0*fac
      end do
      if(constraint)then
       arg=a(offset)**2.d0*x**2.d0
       ex=dexp(-arg)*x**rn(offset)
       fac=fixed_coeff*ex
       y=y+fac
       dyda(offset)=(-2.d0)*a(offset)*x**2.d0*fac
      endif
      if(local)y=y-(znuc*x**rn(offset))
      return
      END SUBROUTINE fgauss2


      SUBROUTINE lmder1(r,y,fcn,m,n,x,fvec,fjac,ldfjac,tol,info,ipvt,
     &                  wa,lwa)
      USE pfit_data
      IMPLICIT NONE
      INTEGER m,n,ldfjac,info,lwa
      INTEGER ipvt(n)
      DOUBLE PRECISION tol
      DOUBLE PRECISION x(n),fvec(m),fjac(ldfjac,n),wa(lwa),r(m),y(m)
      EXTERNAL fcn
      INTEGER maxfev,mode,nfev,njev,nprint
      DOUBLE PRECISION factor,ftol,gtol,xtol,zero
      data factor,zero /1.0d2,0.0d0/
      info = 0
      if (n.le.0.or.m.lt.n.or.ldfjac.lt.m.or.tol.lt.zero
     &    .or.lwa.lt.5*n+m) goto 10
      maxfev = 100*(n + 1)
      ftol = tol
      xtol = tol
      gtol = zero
      mode = 1
      nprint = 0
      call lmder(r,y,fcn,m,n,x,fvec,fjac,ldfjac,ftol,xtol,gtol,maxfev,
     &           wa(1),mode,factor,nprint,info,nfev,njev,ipvt,wa(n+1),
     &           wa(2*n+1),wa(3*n+1),wa(4*n+1),wa(5*n+1))
      if (info .eq. 8) info = 4
10    continue
      return
      END SUBROUTINE lmder1

      SUBROUTINE lmder(r,y,fcn,m,n,x,fvec,fjac,ldfjac,ftol,xtol,gtol,
     &                 maxfev,diag,mode,factor,nprint,info,nfev,njev,
     &                 ipvt,qtf,wa1,wa2,wa3,wa4)
      IMPLICIT NONE
      INTEGER m,n,ldfjac,maxfev,mode,nprint,info,nfev,njev
      INTEGER ipvt(n)
      DOUBLE PRECISION ftol,xtol,gtol,factor
      DOUBLE PRECISION x(n),fvec(m),fjac(ldfjac,n),diag(n),qtf(n),
     &                 wa1(n),wa2(n),wa3(n),wa4(m),r(m),y(m)
      INTEGER i,iflag,iter,j,l
      DOUBLE PRECISION actred,delta,dirder,epsmch,fnorm,fnorm1,gnorm,
     &                 one,par,pnorm,prered,p1,p5,p25,p75,p0001,ratio,
     &                 sum,temp,temp1,temp2,xnorm,zero
      DOUBLE PRECISION dpmpar,enorm
      data one,p1,p5,p25,p75,p0001,zero
     &     /1.0d0,1.0d-1,5.0d-1,2.5d-1,7.5d-1,1.0d-4,0.0d0/
      epsmch = dpmpar(1)
      info = 0
      iflag = 0
      nfev = 0
      njev = 0
      if (n .le. 0 .or. m .lt. n .or. ldfjac .lt. m
     &    .or. ftol .lt. zero .or. xtol .lt. zero .or. gtol .lt. zero
     &    .or. maxfev .le. 0 .or. factor .le. zero) go to 300
      if (mode .ne. 2) go to 20
      do 10 j = 1, n
         if (diag(j) .le. zero) go to 300
   10    continue
   20 continue
      iflag = 1
      call fcn(r,y,m,n,x,fvec,fjac,ldfjac,iflag)
      nfev = 1
      if (iflag .lt. 0) go to 300
      fnorm = enorm(m,fvec)
      par = zero
      iter = 1
   30 continue
         iflag = 2
         call fcn(r,y,m,n,x,fvec,fjac,ldfjac,iflag)
         njev = njev + 1
         if (iflag .lt. 0) go to 300
         if (nprint .le. 0) go to 40
         iflag = 0
         if (mod(iter-1,nprint) .eq. 0)
     &      call fcn(r,y,m,n,x,fvec,fjac,ldfjac,iflag)
         if (iflag .lt. 0) go to 300
   40    continue
         call qrfac(m,n,fjac,ldfjac,.true.,ipvt,n,wa1,wa2,wa3)
         if (iter .ne. 1) go to 80
         if (mode .eq. 2) go to 60
         do 50 j = 1, n
            diag(j) = wa2(j)
            if (wa2(j) .eq. zero) diag(j) = one
   50       continue
   60    continue
         do 70 j = 1, n
            wa3(j) = diag(j)*x(j)
   70       continue
         xnorm = enorm(n,wa3)
         delta = factor*xnorm
         if (delta .eq. zero) delta = factor
   80    continue
         do 90 i = 1, m
            wa4(i) = fvec(i)
   90       continue
         do 130 j = 1, n
            if (fjac(j,j) .eq. zero) go to 120
            sum = zero
            do 100 i = j, m
               sum = sum + fjac(i,j)*wa4(i)
  100          continue
            temp = -sum/fjac(j,j)
            do 110 i = j, m
               wa4(i) = wa4(i) + fjac(i,j)*temp
  110          continue
  120       continue
            fjac(j,j) = wa1(j)
            qtf(j) = wa4(j)
  130       continue
         gnorm = zero
         if (fnorm .eq. zero) go to 170
         do 160 j = 1, n
            l = ipvt(j)
            if (wa2(l) .eq. zero) go to 150
            sum = zero
            do 140 i = 1, j
               sum = sum + fjac(i,j)*(qtf(i)/fnorm)
  140          continue
            gnorm = dmax1(gnorm,dabs(sum/wa2(l)))
  150       continue
  160       continue
  170    continue
         if (gnorm .le. gtol) info = 4
         if (info .ne. 0) go to 300
         if (mode .eq. 2) go to 190
         do 180 j = 1, n
            diag(j) = dmax1(diag(j),wa2(j))
  180       continue
  190    continue
  200    continue
            call lmpar(n,fjac,ldfjac,ipvt,diag,qtf,delta,par,wa1,wa2,
     &                 wa3,wa4)
            do 210 j = 1, n
               wa1(j) = -wa1(j)
               wa2(j) = x(j) + wa1(j)
               wa3(j) = diag(j)*wa1(j)
  210          continue
            pnorm = enorm(n,wa3)
            if (iter .eq. 1) delta = dmin1(delta,pnorm)
            iflag = 1
            call fcn(r,y,m,n,wa2,wa4,fjac,ldfjac,iflag)
            nfev = nfev + 1
            if (iflag .lt. 0) go to 300
            fnorm1 = enorm(m,wa4)
            actred = -one
            if (p1*fnorm1 .lt. fnorm) actred = one - (fnorm1/fnorm)**2
            do 230 j = 1, n
               wa3(j) = zero
               l = ipvt(j)
               temp = wa1(l)
               do 220 i = 1, j
                  wa3(i) = wa3(i) + fjac(i,j)*temp
  220             continue
  230          continue
            temp1 = enorm(n,wa3)/fnorm
            temp2 = (dsqrt(par)*pnorm)/fnorm
            prered = temp1**2 + temp2**2/p5
            dirder = -(temp1**2 + temp2**2)
            ratio = zero
            if (prered .ne. zero) ratio = actred/prered
            if (ratio .gt. p25) go to 240
               if (actred .ge. zero) temp = p5
               if (actred .lt. zero)
     &            temp = p5*dirder/(dirder + p5*actred)
               if (p1*fnorm1 .ge. fnorm .or. temp .lt. p1) temp = p1
               delta = temp*dmin1(delta,pnorm/p1)
               par = par/temp
               go to 260
  240       continue
               if (par .ne. zero .and. ratio .lt. p75) go to 250
               delta = pnorm/p5
               par = p5*par
  250          continue
  260       continue
            if (ratio .lt. p0001) go to 290
            do 270 j = 1, n
               x(j) = wa2(j)
               wa2(j) = diag(j)*x(j)
  270          continue
            do 280 i = 1, m
               fvec(i) = wa4(i)
  280          continue
            xnorm = enorm(n,wa2)
            fnorm = fnorm1
            iter = iter + 1
  290       continue
            if (dabs(actred) .le. ftol .and. prered .le. ftol
     &          .and. p5*ratio .le. one) info = 1
            if (delta .le. xtol*xnorm) info = 2
            if (dabs(actred) .le. ftol .and. prered .le. ftol
     &          .and. p5*ratio .le. one .and. info .eq. 2) info = 3
            if (info .ne. 0) go to 300
            if (nfev .ge. maxfev) info = 5
            if (dabs(actred) .le. epsmch .and. prered .le. epsmch
     &          .and. p5*ratio .le. one) info = 6
            if (delta .le. epsmch*xnorm) info = 7
            if (gnorm .le. epsmch) info = 8
            if (info .ne. 0) go to 300
            if (ratio .lt. p0001) go to 200
         go to 30
  300 continue
      if (iflag .lt. 0) info = iflag
      iflag = 0
      if (nprint .gt. 0) call fcn(r,y,m,n,x,fvec,fjac,ldfjac,iflag)
      return
      END SUBROUTINE lmder

      SUBROUTINE lmpar(n,r,ldr,ipvt,diag,qtb,delta,par,x,sdiag,wa1,
     &                 wa2)
      IMPLICIT NONE
      INTEGER n,ldr
      INTEGER ipvt(n)
      DOUBLE PRECISION delta,par
      DOUBLE PRECISION r(ldr,n),diag(n),qtb(n),x(n),sdiag(n),wa1(n),
     &                 wa2(n)
      INTEGER i,iter,j,jm1,jp1,k,l,nsing
      DOUBLE PRECISION dxnorm,dwarf,fp,gnorm,parc,parl,paru,p1,p001,
     &                 sum,temp,zero
      DOUBLE PRECISION dpmpar,enorm
      data p1,p001,zero /1.0d-1,1.0d-3,0.0d0/
      dwarf = dpmpar(2)
      nsing = n
      do 10 j = 1, n
         wa1(j) = qtb(j)
         if (r(j,j) .eq. zero .and. nsing .eq. n) nsing = j - 1
         if (nsing .lt. n) wa1(j) = zero
   10    continue
      if (nsing .lt. 1) go to 50
      do 40 k = 1, nsing
         j = nsing - k + 1
         wa1(j) = wa1(j)/r(j,j)
         temp = wa1(j)
         jm1 = j - 1
         if (jm1 .lt. 1) go to 30
         do 20 i = 1, jm1
            wa1(i) = wa1(i) - r(i,j)*temp
   20       continue
   30    continue
   40    continue
   50 continue
      do 60 j = 1, n
         l = ipvt(j)
         x(l) = wa1(j)
   60    continue
      iter = 0
      do 70 j = 1, n
         wa2(j) = diag(j)*x(j)
   70    continue
      dxnorm = enorm(n,wa2)
      fp = dxnorm - delta
      if (fp .le. p1*delta) go to 220
      parl = zero
      if (nsing .lt. n) go to 120
      do 80 j = 1, n
         l = ipvt(j)
         wa1(j) = diag(l)*(wa2(l)/dxnorm)
   80    continue
      do 110 j = 1, n
         sum = zero
         jm1 = j - 1
         if (jm1 .lt. 1) go to 100
         do 90 i = 1, jm1
            sum = sum + r(i,j)*wa1(i)
   90       continue
  100    continue
         wa1(j) = (wa1(j) - sum)/r(j,j)
  110    continue
      temp = enorm(n,wa1)
      parl = ((fp/delta)/temp)/temp
  120 continue
      do 140 j = 1, n
         sum = zero
         do 130 i = 1, j
            sum = sum + r(i,j)*qtb(i)
  130       continue
         l = ipvt(j)
         wa1(j) = sum/diag(l)
  140    continue
      gnorm = enorm(n,wa1)
      paru = gnorm/delta
      if (paru .eq. zero) paru = dwarf/dmin1(delta,p1)
      par = dmax1(par,parl)
      par = dmin1(par,paru)
      if (par .eq. zero) par = gnorm/dxnorm
  150 continue
         iter = iter + 1
         if (par .eq. zero) par = dmax1(dwarf,p001*paru)
         temp = dsqrt(par)
         do 160 j = 1, n
            wa1(j) = temp*diag(j)
  160       continue
         call qrsolv(n,r,ldr,ipvt,wa1,qtb,x,sdiag,wa2)
         do 170 j = 1, n
            wa2(j) = diag(j)*x(j)
  170       continue
         dxnorm = enorm(n,wa2)
         temp = fp
         fp = dxnorm - delta
         if (dabs(fp) .le. p1*delta
     &       .or. parl .eq. zero .and. fp .le. temp
     &            .and. temp .lt. zero .or. iter .eq. 10) go to 220
         do 180 j = 1, n
            l = ipvt(j)
            wa1(j) = diag(l)*(wa2(l)/dxnorm)
  180       continue
         do 210 j = 1, n
            wa1(j) = wa1(j)/sdiag(j)
            temp = wa1(j)
            jp1 = j + 1
            if (n .lt. jp1) go to 200
            do 190 i = jp1, n
               wa1(i) = wa1(i) - r(i,j)*temp
  190          continue
  200       continue
  210       continue
         temp = enorm(n,wa1)
         parc = ((fp/delta)/temp)/temp
         if (fp .gt. zero) parl = dmax1(parl,par)
         if (fp .lt. zero) paru = dmin1(paru,par)
         par = dmax1(parl,par+parc)
         go to 150
  220 continue
      if (iter .eq. 0) par = zero
      return
      END SUBROUTINE lmpar

      SUBROUTINE qrfac(m,n,a,lda,pivot,ipvt,lipvt,rdiag,acnorm,wa)
      IMPLICIT NONE
      INTEGER m,n,lda,lipvt
      INTEGER ipvt(lipvt)
      LOGICAL pivot
      DOUBLE PRECISION a(lda,n),rdiag(n),acnorm(n),wa(n)
      INTEGER i,j,jp1,k,kmax,minmn
      DOUBLE PRECISION ajnorm,epsmch,one,p05,sum,temp,zero
      DOUBLE PRECISION dpmpar,enorm
      data one,p05,zero /1.0d0,5.0d-2,0.0d0/
      epsmch = dpmpar(1)
      do 10 j = 1, n
         acnorm(j) = enorm(m,a(1,j))
         rdiag(j) = acnorm(j)
         wa(j) = rdiag(j)
         if (pivot) ipvt(j) = j
   10    continue
      minmn = min0(m,n)
      do 110 j = 1, minmn
         if (.not.pivot) go to 40
         kmax = j
         do 20 k = j, n
            if (rdiag(k) .gt. rdiag(kmax)) kmax = k
   20       continue
         if (kmax .eq. j) go to 40
         do 30 i = 1, m
            temp = a(i,j)
            a(i,j) = a(i,kmax)
            a(i,kmax) = temp
   30       continue
         rdiag(kmax) = rdiag(j)
         wa(kmax) = wa(j)
         k = ipvt(j)
         ipvt(j) = ipvt(kmax)
         ipvt(kmax) = k
   40    continue
         ajnorm = enorm(m-j+1,a(j,j))
         if (ajnorm .eq. zero) go to 100
         if (a(j,j) .lt. zero) ajnorm = -ajnorm
         do 50 i = j, m
            a(i,j) = a(i,j)/ajnorm
   50       continue
         a(j,j) = a(j,j) + one
         jp1 = j + 1
         if (n .lt. jp1) go to 100
         do 90 k = jp1, n
            sum = zero
            do 60 i = j, m
               sum = sum + a(i,j)*a(i,k)
   60          continue
            temp = sum/a(j,j)
            do 70 i = j, m
               a(i,k) = a(i,k) - temp*a(i,j)
   70          continue
            if (.not.pivot .or. rdiag(k) .eq. zero) go to 80
            temp = a(j,k)/rdiag(k)
            rdiag(k) = rdiag(k)*dsqrt(dmax1(zero,one-temp**2))
            if (p05*(rdiag(k)/wa(k))**2 .gt. epsmch) go to 80
            rdiag(k) = enorm(m-j,a(jp1,k))
            wa(k) = rdiag(k)
   80       continue
   90       continue
  100    continue
         rdiag(j) = -ajnorm
  110    continue
      return
      END SUBROUTINE qrfac

      SUBROUTINE qrsolv(n,r,ldr,ipvt,diag,qtb,x,sdiag,wa)
      IMPLICIT NONE
      INTEGER n,ldr
      INTEGER ipvt(n)
      DOUBLE PRECISION r(ldr,n),diag(n),qtb(n),x(n),sdiag(n),wa(n)
      INTEGER i,j,jp1,k,kp1,l,nsing
      DOUBLE PRECISION cos,cotan,p5,p25,qtbpj,sin,sum,tan,temp,zero
      data p5,p25,zero /5.0d-1,2.5d-1,0.0d0/
      do 20 j = 1, n
         do 10 i = j, n
            r(i,j) = r(j,i)
   10       continue
         x(j) = r(j,j)
         wa(j) = qtb(j)
   20    continue
      do 100 j = 1, n
         l = ipvt(j)
         if (diag(l) .eq. zero) go to 90
         do 30 k = j, n
            sdiag(k) = zero
   30       continue
         sdiag(j) = diag(l)
         qtbpj = zero
         do 80 k = j, n
            if (sdiag(k) .eq. zero) go to 70
            if (dabs(r(k,k)) .ge. dabs(sdiag(k))) go to 40
               cotan = r(k,k)/sdiag(k)
               sin = p5/dsqrt(p25+p25*cotan**2)
               cos = sin*cotan
               go to 50
   40       continue
               tan = sdiag(k)/r(k,k)
               cos = p5/dsqrt(p25+p25*tan**2)
               sin = cos*tan
   50       continue
            r(k,k) = cos*r(k,k) + sin*sdiag(k)
            temp = cos*wa(k) + sin*qtbpj
            qtbpj = -sin*wa(k) + cos*qtbpj
            wa(k) = temp
            kp1 = k + 1
            if (n .lt. kp1) go to 70
            do 60 i = kp1, n
               temp = cos*r(i,k) + sin*sdiag(i)
               sdiag(i) = -sin*r(i,k) + cos*sdiag(i)
               r(i,k) = temp
   60          continue
   70       continue
   80       continue
   90    continue
         sdiag(j) = r(j,j)
         r(j,j) = x(j)
  100    continue
      nsing = n
      do 110 j = 1, n
         if (sdiag(j) .eq. zero .and. nsing .eq. n) nsing = j - 1
         if (nsing .lt. n) wa(j) = zero
  110    continue
      if (nsing .lt. 1) go to 150
      do 140 k = 1, nsing
         j = nsing - k + 1
         sum = zero
         jp1 = j + 1
         if (nsing .lt. jp1) go to 130
         do 120 i = jp1, nsing
            sum = sum + r(i,j)*wa(i)
  120       continue
  130    continue
         wa(j) = (wa(j) - sum)/sdiag(j)
  140    continue
  150 continue
      do 160 j = 1, n
         l = ipvt(j)
         x(l) = wa(j)
  160    continue
      return
      END SUBROUTINE qrsolv

      SUBROUTINE coeffs(x,ys,as)
      USE pfit_data
c
c.....fit the coefficients and not the exponents
c
c.....Note this is a kluge. The routines above won't 
c.....fit just a subset of the variable parameters, and I'm 
c.....sufficiently lazy that I can't be bothered to fix it so they do.
c.....The following is a bit of old code I had lying around based
c.....on some poor quality numerical recipes routines, which will fit
c.....the required subset - controlled using the ia vector: 
c.....ia(parameter) = 1 (fit it) = 0 (don't fit it) 
c.....This is actually a full non-linear code but it is sufficiently
c.....terrible that the non-linear stuff doesn't work on a problem
c.....of this complexity. However, the linear fitting is usually
c.....perfectly adequate to provide an initial best fit of the 
c.....coefficients.
c
      IMPLICIT NONE
      INTEGER nca,maxit,ia(ma),i,j,k,n,it
      DOUBLE PRECISION alamda,chisq,alpha(ma,ma),covar(ma,ma),
     & sig(ndata),x(ndata),ys(ndata),as(ma),ochisq,diff,
     & alold,yfits(ndata),mdiff,small,arg,ex,fac
      parameter(maxit=100,small=1.d0)
      LOGICAL one
      CHARACTER*72 dum
      nca=ma
      sig(1:ndata)=1.d0
      alamda=-1.d0
      ochisq=1.d6
      ia(1:offset)=0
      ia(offset+1:ma)=1
      do it=1,maxit
       alold=alamda
       call mrqmin(x,ys,sig,ndata,as,ia,ma,covar,alpha,
     &   nca,chisq,alamda)
       diff=ochisq-chisq
       ochisq=chisq
       if(diff.lt.small.and.alamda.lt.alold)goto 1
      end do ! iterations
      write(6,*)
     &'too many iterations in initial fitting of coefficients'
      stop
1     write(6,*)
     &'Sum of squared deviations after initial coefficient tweak: ',diff
      write(6,*)' '
      return
      END SUBROUTINE coeffs

      SUBROUTINE mrqmin(x,y,sig,ndata,a,ia,ma,covar,alpha,
     &   nca,chisq,alamda)
      IMPLICIT NONE
      INTEGER ma,nca,ia(ma),mmax,ndata
      DOUBLE PRECISION alamda,chisq,a(ma),alpha(nca,nca),covar(nca,nca),
     & sig(ndata),x(ndata),y(ndata)
      parameter (mmax=100)
      INTEGER j,k,l,m,mfit
      DOUBLE PRECISION ochisq,atry(mmax),beta(mmax),da(mmax)
      save ochisq,atry,beta,da,mfit
      if(alamda.lt.0.)then
       mfit=0
       do 11 j=1,ma
        if(ia(j).ne.0)mfit=mfit+1
11     continue 
       alamda=0.001d0
       call mrqcof(x,y,sig,ndata,a,ia,ma,alpha,beta,nca,chisq)
       ochisq=chisq
       do 12 j=1,ma
         atry(j)=a(j)
12     continue
      endif
      j=0
      do 14 l=1,ma
       if(ia(l).ne.0)then
        j=j+1
        k=0
        do 13 m=1,ma
         if(ia(m).ne.0)then
          k=k+1
          covar(j,k)=alpha(j,k)
         endif
13      continue
        covar(j,j)=alpha(j,j)*(1.+alamda)
        da(j)=beta(j)
       endif
14    continue
      call gaussj(covar,mfit,nca,da,1,1)
      if(alamda.eq.0.)then
       call covsrt(covar,nca,ma,ia,mfit)
       return
      endif
      j=0
      do 15 l=1,ma   
       if(ia(l).ne.0)then
        j=j+1
        atry(l)=atry(l)+da(j)
       endif
15    continue
      call mrqcof(x,y,sig,ndata,atry,ia,ma,covar,da,nca,chisq)
      if(chisq.lt.ochisq)then
       alamda=0.1d0*alamda
       ochisq=chisq
       j=0
       do 17 l=1,ma
        if(ia(l).ne.0)then
         j=j+1
         k=0
         do 16 m=1,ma  
          if(ia(m).ne.0)then
           k=k+1
           alpha(j,k)=covar(j,k)
          endif
16       continue
         beta(j)=da(j)
         a(l)=atry(l)
        endif
17     continue
      else
        alamda=10.d0*alamda
        chisq=ochisq
      endif
      return
      END SUBROUTINE mrqmin

      SUBROUTINE covsrt(covar,npc,ma,ia,mfit)
      IMPLICIT NONE
      INTEGER ma,mfit,npc,ia(ma)
      DOUBLE PRECISION covar(npc,npc)
      INTEGER i,j,k
      DOUBLE PRECISION swap
      do 12 i=mfit+1,ma
       do 11 j=1,i
        covar(i,j)=0.
        covar(j,i)=0.
11     continue
12    continue
      k=mfit
      do 15 j=ma,1,-1
       if(ia(j).ne.0)then
        do 13 i=1,ma
         swap=covar(i,k)
         covar(i,k)=covar(i,j)
         covar(i,j)=swap
13      continue
        do 14 i=1,ma
         swap=covar(k,i)
         covar(k,i)=covar(j,i)
         covar(j,i)=swap
14      continue
        k=k-1
       endif
15    continue
      return
      END SUBROUTINE covsrt


      SUBROUTINE mrqcof(x,y,sig,ndata,a,ia,ma,alpha,beta,nalp,ch
     &isq)
      IMPLICIT NONE
      INTEGER ma,nalp,ia(ma),ndata
      DOUBLE PRECISION chisq,a(ma),alpha(nalp,nalp),beta(ma),sig(ndata),
     & x(ndata),y(ndata)
      INTEGER mfit,i,j,k,l,m
      DOUBLE PRECISION dy,sig2i,wt,ymod,dyda(ma)
      mfit=0
      do 11 j=1,ma
       if(ia(j).ne.0)mfit=mfit+1
11    continue
      do 13 j=1,mfit
       do 12 k=1,j
        alpha(j,k)=0.
12     continue
       beta(j)=0.
13    continue
      chisq=0.
      do 16 i=1,ndata
        if(x(i).lt.1.d-15)x(i)=x(i+1) 
        call fgauss2(x(i),a,ymod,dyda)
        sig2i=1./(sig(i)*sig(i))
        dy=y(i)-ymod
        do j=1,ma
        enddo
        j=0
        do 15 l=1,ma
         if(ia(l).ne.0)then
          j=j+1
          wt=dyda(l)*sig2i
          k=0
          do 14 m=1,l
           if(ia(m).ne.0)then
            k=k+1
            alpha(j,k)=alpha(j,k)+wt*dyda(m)
           endif
14        continue
          beta(j)=beta(j)+dy*wt
         endif
15      continue
        chisq=chisq+dy*dy*sig2i
16    continue
      do 18 j=2,mfit
        do 17 k=1,j-1
          alpha(k,j)=alpha(j,k)
17      continue
18    continue
      return
      END SUBROUTINE mrqcof

      SUBROUTINE gaussj(a,n,np,b,m,mp)
      IMPLICIT NONE
      INTEGER m,mp,n,np,nmax
      parameter (nmax=500)
      INTEGER i,icol,irow,j,k,l,ll,indxc(nmax),indxr(nmax),
     &ipiv(nmax)
      DOUBLE PRECISION big,dum,pivinv
      DOUBLE PRECISION a(np,np),b(np,mp)
      if(n.gt.nmax)then
       write(6,*)'n greater than nmax in gaussj.'
       write(6,*)'adjust parameter sizes.'
       stop
      endif
      do 11 j=1,n
        ipiv(j)=0
11    continue
      do 22 i=1,n
        big=0.
        do 13 j=1,n
          if(ipiv(j).ne.1)then
            do 12 k=1,n
              if (ipiv(k).eq.0) then
                if (abs(a(j,k)).ge.big)then
                  big=abs(a(j,k))
                  irow=j
                  icol=k
                endif
              else if (ipiv(k).gt.1) then
                write(6,*) 'singular matrix in gaussj - test 1'
                stop
              endif
12          continue
          endif
13      continue
        ipiv(icol)=ipiv(icol)+1
        if (irow.ne.icol) then
          do 14 l=1,n
            dum=a(irow,l)
            a(irow,l)=a(icol,l)
            a(icol,l)=dum
14        continue
          do 15 l=1,m
            dum=b(irow,l)
            b(irow,l)=b(icol,l)
            b(icol,l)=dum
15        continue
        endif
        indxr(i)=irow
        indxc(i)=icol
        if (a(icol,icol).eq.0.) then
         write(6,*) 'singular matrix in gaussj - test 2.'
         stop
        endif
        pivinv=1./a(icol,icol)
        a(icol,icol)=1.
        do 16 l=1,n
          a(icol,l)=a(icol,l)*pivinv
16      continue
        do 17 l=1,m
          b(icol,l)=b(icol,l)*pivinv
17      continue
        do 21 ll=1,n
          if(ll.ne.icol)then
            dum=a(ll,icol)
            a(ll,icol)=0.
            do 18 l=1,n
              a(ll,l)=a(ll,l)-a(icol,l)*dum
18          continue
            do 19 l=1,m
              b(ll,l)=b(ll,l)-b(icol,l)*dum
19          continue
          endif
21      continue
22    continue
      do 24 l=n,1,-1
        if(indxr(l).ne.indxc(l))then
          do 23 k=1,n
            dum=a(k,indxr(l))
            a(k,indxr(l))=a(k,indxc(l))
            a(k,indxc(l))=dum
23        continue
        endif
24    continue
      return
      END SUBROUTINE gaussj

      SUBROUTINE reason(info,tol)
      IMPLICIT NONE
      INTEGER info
      DOUBLE PRECISION tol
      if(info.eq.1)then
      write(6,*)'Relative error in the sum of squares is less than TOL.'
      write(6,*)'Value of TOL: ',tol
      elseif(info.eq.2)then
      write(6,*)
     &'The relative error between the parameter vector and the solution'
      write(6,*)'vector is less than TOL. Value of TOL: ',tol
      elseif(info.eq.3)then
      write(6,*)
     &'The relative error in the sum of squares is less than TOL.'
      write(6,*)
     &'The relative error between the parameter vector and the'
      write(6,*)'solution vector is also less than TOL.'
      write(6,*)'Value of TOL: ',tol
      endif
      write(6,*)' '
      return
      END SUBROUTINE reason

      SUBROUTINE errmdt(info,tol)
      IMPLICIT NONE
      INTEGER info
      DOUBLE PRECISION tol
      if(info.eq.0)info=1
      goto(1,4,5,6,7),info
1     write(6,*)'Improper input parameters.'
      return
4     write(6,*)'Fvec is orthogonal to the columns of the jacobian to '
      write(6,*)'machine precision.'
      return
5     write(6,*)'Number of calls to fgauss for function computation'
      write(6,*)'exceeds 100*(n+1), where n is the number of variable '
      write(6,*)'parameters.'
      return
6     write(6,*)'TOL is too small. no further reduction in the sum of '
      write(6,*)'squared deviations is possible.'
      return
7     write(6,*)'TOL is too small. no further improvement in the'
      write(6,*)' approximate solution vector x is possible.'
      write(6,*)' '
      return
      END SUBROUTINE errmdt

      DOUBLE PRECISION function dpmpar(i)
      IMPLICIT NONE
      INTEGER i
      DOUBLE PRECISION dmach(3)
c.....floating point parameters derived for dec alpha 
c.....(mdt 1/6/97)
c.....machine precision:
      data dmach(1) /2.22044604926d-16/
c.....smallest meaningful number:
      data dmach(2) /2.22507385852d-308/
c.....largest meaningful number:
      data dmach(3) /1.79769313485d+308/
      dpmpar = dmach(i)
      return
      END FUNCTION dpmpar

      DOUBLE PRECISION function enorm(n,x)
      IMPLICIT NONE
      INTEGER n
      DOUBLE PRECISION x(n)
      INTEGER i
      DOUBLE PRECISION agiant,floatn,one,rdwarf,rgiant,s1,s2,s3,xabs,
     &                 x1max,x3max,zero
      data one,zero,rdwarf,rgiant /1.0d0,0.0d0,3.834d-20,1.304d19/
      s1 = zero
      s2 = zero
      s3 = zero
      x1max = zero
      x3max = zero
      floatn = n
      agiant = rgiant/floatn
      do 90 i = 1, n
         xabs = dabs(x(i))
         if (xabs .gt. rdwarf .and. xabs .lt. agiant) go to 70
            if (xabs .le. rdwarf) go to 30
               if (xabs .le. x1max) go to 10
                  s1 = one + s1*(x1max/xabs)**2
                  x1max = xabs
                  go to 20
10          continue
                  s1 = s1 + (xabs/x1max)**2
20          continue
              go to 60
30       continue
               if (xabs .le. x3max) go to 40
                  s3 = one + s3*(x3max/xabs)**2
                  x3max = xabs
                  go to 50
40          continue
                  if (xabs .ne. zero) s3 = s3 + (xabs/x3max)**2
50          continue
60       continue
           go to 80
70    continue
           s2 = s2 + xabs**2
80    continue
90    continue
      if (s1 .eq. zero) go to 100
         enorm = x1max*dsqrt(s1+(s2/x1max)/x1max)
         go to 130
100    continue
         if (s2 .eq. zero) go to 110
            if (s2 .ge. x3max)
     &         enorm = dsqrt(s2*(one+(x3max/s2)*(x3max*s3)))
            if (s2 .lt. x3max)
     &         enorm = dsqrt(x3max*((s2/x3max)+(x3max*s3)))
            go to 120
110      continue
            enorm = x3max*dsqrt(s3)
120    continue
130   continue
      return
      END FUNCTION enorm
